from pathlib import Path

import pandas as pd
import numpy as np
from chardet.universaldetector import UniversalDetector

from .common import OED_TYPE_TO_NAME, OdsException, PANDAS_COMPRESSION_MAP, PANDAS_DEFAULT_NULL_VALUES, is_relative
from .forex import convert_currency
from .schema import OedSchema

try:
    from functools import cached_property
except ImportError:  # support for python < 3.8
    _missing = object()

    class cached_property(object):
        """A decorator that converts a function into a lazy property.  The
        function wrapped is called the first time to retrieve the result
        and then that calculated result is used the next time you access
        the value::

            class Foo(object):

                @cached_property
                def foo(self):
                    # calculate something important here
                    return 42

        The class has to have a `__dict__` in order for this property to
        work.
        """

        # implementation detail: this property is implemented as non-data
        # descriptor.  non-data descriptors are only invoked if there is
        # no entry with the same name in the instance's __dict__.
        # this allows us to completely get rid of the access function call
        # overhead.  If one choses to invoke __get__ by hand the property
        # will still work as expected because the lookup logic is replicated
        # in __get__ for manual invocation.

        def __init__(self, func, name=None, doc=None):
            self.__name__ = name or func.__name__
            self.__module__ = func.__module__
            self.__doc__ = doc or func.__doc__
            self.func = func

        def __get__(self, obj, type=None):
            if obj is None:
                return self
            value = obj.__dict__.get(self.__name__, _missing)
            if value is _missing:
                value = self.func(obj)
                obj.__dict__[self.__name__] = value
            return value


def detect_encoding(filepath):
    """
    Given a path to a CSV of unknown encoding
    read lines to detect its encoding type

    :param filepath: Filepath to check
    :type  filepath: str

    :return: Example `{'encoding': 'ISO-8859-1', 'confidence': 0.73, 'language': ''}`
    :rtype: dict
    """
    detector = UniversalDetector()
    with open(filepath, 'rb') as f:
        for line in f:
            detector.feed(line)
            if detector.done:
                break
    detector.close()
    return detector.result


class OedSource:
    """
    Class to represent and manage oed source (location, account, ...)
    """

    def __init__(self, exposure, oed_type, cur_version_name, sources):
        """

        Args:
            exposure (OedExposure): exposure the OED source is part of
            oed_type (str): OED type of the source
            cur_version_name (str): name of the current version
            sources (dict): all the version/source of the OED source
        """
        self.exposure = exposure
        self.oed_type = oed_type
        self.oed_name = OED_TYPE_TO_NAME[oed_type]
        self.cur_version_name = cur_version_name
        self.sources = sources
        self.loaded = False

    def __str__(self):
        """
        Returns:
            string of the current version source
        """
        return str(self.sources[self.cur_version_name])

    @property
    def info(self):
        """
        information of source to be able to trace its different version and reload it if necessary
        Returns:
            info dict
        """
        return {'cur_version_name': self.cur_version_name, 'sources': self.sources}

    @classmethod
    def from_oed_info(cls, exposure, oed_type: str, oed_info):
        """
        Convert data in oed_info to an OedSource

        Args:
            exposure (OedExposure): Exposure the source is part of
            oed_type (str): type of oed file (Loc, Acc, ...)
            oed_info: info to create OedSource, can be:
                - str for filepath
                - dict for more complex config,
                - OedSource same object is returned
                - DataFrame no source file just oed data is path and can be saved after

        Returns:
            OedSource (or None if  oed_info is None)
        """
        if isinstance(oed_info, (str, Path)):
            return OedSource.from_filepath(exposure, oed_type, filepath=oed_info)
        elif isinstance(oed_info, dict):
            return OedSource(exposure, oed_type, **oed_info)
        elif isinstance(oed_info, OedSource):
            return oed_info
        elif isinstance(oed_info, pd.DataFrame):
            return OedSource.from_dataframe(exposure, oed_type, oed_info)
        elif oed_info is None:
            return None
        else:
            raise OdsException(f'{oed_info} is not a supported format to convert to OedSource')

    @classmethod
    def from_filepath(cls, exposure, oed_type, filepath):
        """
        OedSource Constructor from a filepath
        Args:
            exposure (OedExposure): Exposure the oed source is part of
            oed_type (str): type of file (Loc, Acc, ..)
            filepath (str): path to the oed source file

        Returns:
            OedSource
        """
        return cls(exposure, oed_type, 'orig', {'orig': {'source_type': 'filepath', 'filepath': filepath}})

    @classmethod
    def prepare_df(cls, df, column_to_field, ods_fields):
        """
        Complete the Oed Dataframe with default valued and required column
        Args:
            df: oed dataframe
            column_to_field: dict mapping column to their field info
            ods_fields: the ods_field info for this oed source type

        Returns:
            df
        """
        # set default values
        for col, field_info in column_to_field.items():
            field_info = column_to_field.get(col)
            if (field_info
                    and field_info['Default'] != 'n/a'
                    and (df[col].isna().any() or (field_info['pd_dtype'] == 'category' and df[col].isnull().any()))):
                if field_info['pd_dtype'] == 'category':
                    df[col] = df[col].cat.add_categories(field_info['Default']).fillna(field_info['Default'])
                else:
                    df[col].fillna(df[col].dtype.type(field_info['Default']), inplace=True)

        # add required columns that allow blank values if missing
        present_field = set(field_info['Input Field Name'] for field_info in column_to_field.values())
        for field_info in ods_fields.values():
            col = field_info['Input Field Name']
            if col not in present_field:
                if field_info.get('Required Field') == 'R' and field_info.get("Allow blanks?").upper() == "YES":
                    if field_info['pd_dtype'] == 'category':
                        df[col] = '' if field_info['Default'] == 'n/a' else field_info['Default']
                        df[col] = df[col].astype('category')
                    else:
                        df[col] = np.nan
                        df[col] = df[col].astype(field_info['pd_dtype'])
                        if field_info['Default'] != 'n/a':
                            df[col] = df[col].fillna(df[col].dtype.type(field_info['Default'])).astype(field_info['pd_dtype'])
        return df

    @classmethod
    def from_dataframe(cls, exposure, oed_type, oed_df: pd.DataFrame):
        """
        OedSource Constructor from a filepath
        Args:
            exposure (OedExposure): Exposure the oed source is part of
            oed_type (str): type of file (Loc, Acc, ..)
            oed_df (pd.DataFrame): DataFrame that represent the Oed Source

        Returns:
            OedSource
        """
        oed_source = cls(exposure, oed_type, 'orig', {'orig': {'source_type': 'DataFrame'}})
        ods_fields = exposure.get_input_fields(oed_type)
        pd_dtype = {}
        to_tmp_dtype = {}
        column_to_field = OedSchema.column_to_field(oed_df.columns, ods_fields)
        for column in oed_df.columns:
            if column in column_to_field:
                pd_dtype[column] = column_to_field[column]['pd_dtype']
            else:
                pd_dtype[column] = 'category'
            if pd_dtype[column] == 'category':  # we need to convert to str first
                to_tmp_dtype[column] = 'str'
            elif pd_dtype[column].startswith('Int'):
                to_tmp_dtype[column] = 'float'

        oed_df = oed_df.astype(to_tmp_dtype).astype(pd_dtype)
        if exposure.use_field:
            oed_df = OedSchema.use_field(oed_df, ods_fields)
        oed_source.dataframe = cls.prepare_df(oed_df, column_to_field, ods_fields)
        oed_source.loaded = True
        return oed_source

    @cached_property
    def dataframe(self):
        """Dataframe view of the OedSource, loaded once"""
        self.loaded = True
        df = self.load_dataframe()
        if self.exposure.use_field:
            df = OedSchema.use_field(df, self.exposure.get_input_fields(self.oed_type))
        return df

    @property
    def current_source(self):
        """
        current version of the oed source
        Returns:
            source dict of the current version
        """
        return self.sources[self.cur_version_name]

    def get_input_fields(self):
        """
        Returns:
            OED schema input field definition
        """
        return self.exposure.get_input_fields(self.oed_type)

    def get_column_to_field(self):
        """
        Returns:
            mapping between column in dataframe and field definition
        """
        return OedSchema.column_to_field(
            self.dataframe.columns,
            self.get_input_fields()
        )

    def load_dataframe(self, version_name=None):
        """
        load the dataframe from a version of oed source

        Args:
            version_name (str): name of the version in sources

        Returns:
            Dataframe representing the oed source (pd.DataFrame)
        """
        if version_name is None:
            version_name = self.cur_version_name
        source = self.sources[version_name]
        if source['source_type'] == 'filepath':
            filepath = source['filepath']
            if is_relative(filepath):
                filepath = Path(self.exposure.working_dir, filepath)
            extension = PANDAS_COMPRESSION_MAP.get(source.get('extention')) or Path(filepath).suffix
            if extension == '.parquet':
                oed_df = pd.read_parquet(filepath, **source.get('read_param', {}))
            else:  # default we assume it is csv like
                read_params = {'keep_default_na': False,
                               'na_values': PANDAS_DEFAULT_NULL_VALUES.difference('NA')}
                read_params.update(source.get('read_param', {}))
                oed_df = self.read_csv(filepath, self.exposure.get_input_fields(self.oed_type), **read_params)
        else:
            raise Exception(f"Source type {source['source_type']} is not supported")

        if self.exposure.reporting_currency:
            convert_currency(oed_df,
                             self.oed_type,
                             self.exposure.reporting_currency,
                             self.exposure.currency_conversion,
                             self.exposure.oed_schema)
        return oed_df

    def convert_currency(self):
        """
        convert currency values current of the oed source to exposure.reporting_currency
        """
        if self.loaded:
            convert_currency(self.dataframe,
                             self.oed_type,
                             self.exposure.reporting_currency,
                             self.exposure.currency_conversion,
                             self.exposure.oed_schema)

    def save(self, version_name, source):
        """
        save dataframe as version_name in source
        Args:
            version_name (str): name of the version
            source: str or dict with information to save the dataframe
                str : output path
                dict : {'source_type': 'filepath' # only support for the moment
                        'extension': 'parquet' or all pandas supported extension
                        'write_param' : all args you may want to pass to the pandas writer function (to_parquet, to_csv)

        """
        if isinstance(source, (str, Path)):
            source = {'source_type': 'filepath',
                      'filepath': source,
                      }
        if source['source_type'] == 'filepath':
            filepath = source['filepath']
            if is_relative(filepath):
                filepath = Path(self.exposure.working_dir, filepath)
            Path(filepath).parents[0].mkdir(parents=True, exist_ok=True)
            extension = source.get('extension') or ''.join(Path(filepath).suffixes)
            if extension == 'parquet':
                self.dataframe.to_parquet(filepath, **source.get('write_param', {}))
            else:
                write_param = {'index': False}
                write_param.update(source.get('write_param', {}))
                self.dataframe.to_csv(filepath, **write_param)
        else:
            raise Exception(f"Source type {source['source_type']} is not supported")
        self.cur_version_name = version_name
        self.sources[version_name] = source

    @classmethod
    def read_csv(cls, filepath, ods_fields, df_engine=pd, **kwargs):
        """
        the function read_csv will load a csv file as a DataFrame
        with all the columns converted to the correct dtype and having the correct default.
        it will also try to save space by converting string dtype into categories

        By default, it uses pandas to create the DataFrame. In that case you will need to have pandas installed.
        You can use other options such as Dask or modin using the parameter df_engine.

        Args:
            filepath (str): path to the csv file
            df_engine: engine that will convert csv to a dataframe object (default to pandas if installed)
            kwargs: extra argument that will be passed to the df_engine
        Returns:
            df_engine dataframe of the file with correct dtype and default

        Raises:


        """

        def read_or_try_encoding_read(df_engine, filepath, **read_kwargs):
            #  try to read, if fail try to detect the encoding and update the top function kwargs for future read
            try:
                return df_engine.read_csv(filepath, **kwargs)
            except UnicodeDecodeError as e:
                detected_encoding = detect_encoding(filepath)['encoding']
                if not read_kwargs.get('encoding') and detected_encoding:
                    kwargs['encoding'] = detected_encoding
                    read_kwargs.pop('encoding')
                    return df_engine.read_csv(filepath, encoding=detected_encoding, **read_kwargs)
                else:
                    raise

        if df_engine is None:
            raise Exception("df_engine parameter not specified, you must install pandas"
                            " or pass your DataFrame engine (modin, dask,...)")

        if Path(filepath).suffix == '.gzip' or kwargs.get('compression') == 'gzip':
            # support for gzip  https://stackoverflow.com/questions/60460814/pandas-read-csv-failing-on-gzipped-file-with-unicodedecodeerror-utf-8-codec-c
            with open(filepath, 'rb') as f:
                header = df_engine.read(f, compression='gzip', nrows=0, index_col=False,
                                        encoding=kwargs.get('encoding')).columns
            kwargs['compression'] = 'gzip'
        else:
            header = read_or_try_encoding_read(df_engine, filepath, nrows=0, index_col=False,
                                               encoding=kwargs.get('encoding')).columns

        # match header column name to oed field name and prepare pd_dtype used to read the data
        pd_dtype = {}
        column_to_field = OedSchema.column_to_field(header, ods_fields)
        for col in header:
            if col in column_to_field:
                field_info = column_to_field[col]
                pd_dtype[col] = field_info['pd_dtype']
            else:
                pd_dtype[col] = 'category'

        # read the oed file
        if kwargs.get('compression') == 'gzip':
            with open(filepath, 'rb') as f:
                df = df_engine.read_csv(f, dtype=pd_dtype, **kwargs)
        else:
            df = df_engine.read_csv(filepath, dtype=pd_dtype, **kwargs)

        return cls.prepare_df(df, column_to_field, ods_fields)



import json
import os
from pathlib import Path
import pandas as pd

from .common import OdsException


class OedSchema:
    """
    Object managing information about a certain OED schema

    Attributes:
        schema (dict): information about the OED schema
        json_path (Path or str): path to the json file from where the schema was loaded
    """
    DEFAULT_ODS_SCHEMA_FILE = 'OpenExposureData_Spec.json'
    DEFAULT_ODS_SCHEMA_PATH = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
                                           'data', DEFAULT_ODS_SCHEMA_FILE)

    def __init__(self, schema, json_path):
        """
        Args:
            schema (dict): information about the OED schema
            json_path (Path or str): path to the json file from where the schema was loaded
        """
        self.schema = schema
        self.json_path = json_path

    @property
    def info(self):
        """
        Information necessary to reload the same OedSchema
        """
        return self.json_path

    @classmethod
    def from_oed_schema_info(cls, oed_schema_info):
        """
        Args:
            oed_schema_info: info to create OedSchema object, can be:
                - str to specify your own OedSchema path
                - OedSchema object will be return directly
                - None to get the current version Oed schema

        Returns:
            OedSchema
        """
        if isinstance(oed_schema_info, (str, Path)):
            return cls.from_json(oed_schema_info)
        elif isinstance(oed_schema_info, cls):
            return oed_schema_info
        elif oed_schema_info is None:
            return cls.from_json(cls.DEFAULT_ODS_SCHEMA_PATH)
        else:
            raise OdsException(f"{oed_schema_info} is not a supported format to create {cls} object")

    @classmethod
    def from_json(cls, oed_json):
        """
        Create OedSchema from json file
        Args:
            oed_json (str): OED schema json path

        Returns:
            OedSchema
        """
        with open(oed_json) as f:
            schema = json.load(f)
            # reformat area code for efficient check
            country_area = set()
            for country, areas in schema['area'].items():
                for area in areas:
                    country_area.add((country, area))
            schema["country_area"] = country_area

            return cls(schema, oed_json)

    @staticmethod
    def column_to_field(columns: list, oed_fields: dict, use_generic_flexi=True):
        """
        map column to OED field

        Args:
            columns (list): name of the columns in OED file
            oed_fields (dict): all the field in ods_schema
            use_generic_flexi (bool): if true flexi column return the oed_schema name
                                      if false flexi column are just lower cased
        Return:
            dict mapping between exact OED column name and field name in oed_schema
        """
        # support for name different from standard OED column name
        aliases = {field_info.get('alias').lower(): field_name for field_name, field_info in oed_fields.items() if field_info.get('alias')}
        result = {}
        for column in columns:
            if column.lower() in oed_fields:
                result[column] = oed_fields[column.lower()]
            elif column.lower() in aliases:
                result[column] = oed_fields[aliases[column.lower()]]
            else:
                for field_suffix in ['xx', 'zzz']:
                    for i in range(1, len(column)):
                        field_name = column.lower()[:-i] + field_suffix
                        if field_name in oed_fields:
                            if use_generic_flexi:
                                result[column] = oed_fields[field_name]
                            break
                    else:
                        continue
                    break
                else:
                    pass  # unrecognized/unknown columns are not added
        return result

    @staticmethod
    def use_field(dataframe, ods_fields):
        """
        rename the column in the dataframe to their OED field name

        Args:
            dataframe (pd.DataFrame): dataframe with OED column
            ods_fields (dict): OedSchema input field definition

        Returns:
            dataframe with renamed column
        """
        mapping = {column: field['Input Field Name']
                   for column, field in OedSchema.column_to_field(dataframe.columns, ods_fields, use_generic_flexi=False).items()}
        return dataframe.rename(columns=mapping)

    @staticmethod
    def is_valid_value(value, valid_ranges, allow_blanks):
        """
        >>> OedSchema.is_valid_value(2, [{'min': 0, 'max': 5}, {'min': 10}, {'max': -5}, {'enum': [-999]}])
        True

        >>> OedSchema.is_valid_value(7, [{'min': 0, 'max': 5}, {'min': 10}, {'max': -5}, {'enum': [-999]}])
        False

        >>> OedSchema.is_valid_value(-10, [{'min': 0, 'max': 5}, {'min': 10}, {'max': -5}, {'enum': [-999]}])
        True

        >>> OedSchema.is_valid_value(-999, [{'min': 0, 'max': 5}, {'min': 10}, {'enum': [-999]}])
        True

        Args:
            value: value to test
            valid_ranges (list): list of valid range
            allow_blanks (bool): if True blank value (ie pd.nan) will be considered valid

        Returns:
            True if value is in one of the range
        """

        if pd.isna(value):
            return allow_blanks
        for valid_range in valid_ranges:
            if valid_range.get('min') is not None:
                if value < valid_range.get('min'):
                    continue
            if valid_range.get('max') is not None:
                if value > valid_range.get('max'):
                    continue
            if valid_range.get('enum') is not None:
                if value not in valid_range.get('enum'):
                    continue
            return True
        return False



from .exposure import OedExposure
from .schema import OedSchema
from .source import OedSource
from .common import (
    OdsException, PANDAS_COMPRESSION_MAP, PANDAS_DEFAULT_NULL_VALUES, USUAL_FILE_NAME, OED_TYPE_TO_NAME,
    OED_NAME_TO_TYPE, OED_IDENTIFIER_FIELDS, VALIDATOR_ON_ERROR_ACTION, DEFAULT_VALIDATION_CONFIG, OED_PERIL_COLUMNS
)


__all__ = [
    'OedExposure', 'OedSchema', 'OedSource',
    'OdsException', 'PANDAS_COMPRESSION_MAP', 'PANDAS_DEFAULT_NULL_VALUES', 'USUAL_FILE_NAME', 'OED_TYPE_TO_NAME',
    'OED_NAME_TO_TYPE', 'OED_IDENTIFIER_FIELDS', 'VALIDATOR_ON_ERROR_ACTION', 'DEFAULT_VALIDATION_CONFIG', 'OED_PERIL_COLUMNS',
]



"""
This package manage:
    loading an OED schema
    the loading, checking and saving of OED files
    support conversion of OED to other currencies
"""

import json
from pathlib import Path

from .common import (PANDAS_COMPRESSION_MAP,
                     USUAL_FILE_NAME, OED_TYPE_TO_NAME)
from .schema import OedSchema
from .source import OedSource
from .validator import Validator
from .forex import create_currency_rates


class OedExposure:
    """
    Object grouping all the OED files related to the exposure Data (location, acoount, ri_info, ri_scope)
    and the OED schema to follow
    """
    DEFAULT_EXPOSURE_CONFIG_NAME = 'exposure_info.json'

    def __init__(self,
                 location=None,
                 account=None,
                 ri_info=None,
                 ri_scope=None,
                 oed_schema_info=None,
                 currency_conversion=None,
                 reporting_currency=None,
                 check_oed=False,
                 use_field=False,
                 validation_config=None,
                 working_dir=None):
        """
        Create an OED object,
        each input can be the object itself or  information that will be used to create the object

        Args:
            location (path or dict or OedSource or pd.DataFrame): info for location
            account (path or dict or OedSource or pd.DataFrame): info for account
            ri_info (path or dict or OedSource or pd.DataFrame): info for ri_info
            ri_scope (path or dict or OedSource or pd.DataFrame): info for ri_scope
            oed_schema_info (path_to_json, OedSchema, None): info for oed_schema
            currency_conversion (path_to_json or dict or None): info  currency_conversion
            reporting_currency (str): currency to convert
            check_oed (bool): check if OED files are valid or not
            use_field (bool): if true column name are converted to OED field name on load
        """
        self.use_field = use_field
        self.oed_schema = OedSchema.from_oed_schema_info(oed_schema_info)

        self.location = OedSource.from_oed_info(exposure=self, oed_type='Loc', oed_info=location)
        self.account = OedSource.from_oed_info(exposure=self, oed_type='Acc', oed_info=account)
        self.ri_info = OedSource.from_oed_info(exposure=self, oed_type='ReinsInfo', oed_info=ri_info)
        self.ri_scope = OedSource.from_oed_info(exposure=self, oed_type='ReinsScope', oed_info=ri_scope)

        self.currency_conversion = create_currency_rates(currency_conversion)

        self.reporting_currency = reporting_currency

        self.validation_config = validation_config

        self.working_dir = working_dir

        if check_oed:
            self.check()

    @classmethod
    def from_config(cls, config_fp, **kwargs):
        """
        create an OedExposure from filepath to a config json file
        Args:
            config_fp (str): path
        Returns:
            OedExposure object
        """
        with open(config_fp) as config:
            kwargs['working_dir'] = Path(config_fp).parent
            return cls(**{**json.load(config), **kwargs})

    @classmethod
    def from_dir(cls, oed_dir, **kwargs):
        """
        create an OedExposure from directory path
        Args:
            oed_dir (str): path to the directory
        Returns:
            OedExposure object
        """
        def find_fp(names):
            for name in names:
                for extension in PANDAS_COMPRESSION_MAP.values():
                    if Path(oed_dir, name).with_suffix(extension).is_file():
                        return Path(oed_dir, name).with_suffix(extension)

        if Path(oed_dir, cls.DEFAULT_EXPOSURE_CONFIG_NAME).is_file():
            return cls.from_config(Path(oed_dir, cls.DEFAULT_EXPOSURE_CONFIG_NAME), **kwargs)
        else:
            config = {}
            for attr, filenames in USUAL_FILE_NAME.items():
                config[attr] = find_fp(filenames)

            if Path(oed_dir, OedSchema.DEFAULT_ODS_SCHEMA_FILE).is_file():
                config['oed_schema_info'] = Path(oed_dir, OedSchema.DEFAULT_ODS_SCHEMA_FILE)
            kwargs['working_dir'] = oed_dir
            return cls(**{**config, **kwargs})

    @property
    def info(self):
        """info on exposure Data to be able to reload it"""
        info = {}
        for source in self.get_oed_sources():
            info[source.oed_name] = source.info
        info['oed_schema_info'] = self.oed_schema.info
        return info

    def get_input_fields(self, oed_type):
        """
        helper function to get OED input field info for a particular oed_type
        Args:
            oed_type:

        Returns:
            dict of OED input field info
        """
        return self.oed_schema.schema['input_fields'][oed_type]

    @property
    def reporting_currency(self):
        return self._reporting_currency

    @reporting_currency.setter
    def reporting_currency(self, reporting_currency):
        """setter for reporting_currency that will automatically convert currency in all OED files"""
        self._reporting_currency = reporting_currency
        if reporting_currency:
            for oed_source in self.get_oed_sources():
                oed_source.convert_currency()

    def get_oed_sources(self):
        """
        yield all the non-null oed_sources
        Returns:
            generator of all oed_sources
        """
        for oed_type, oed_name in OED_TYPE_TO_NAME.items():
            oed_source: OedSource = getattr(self, oed_name)
            if oed_source:
                yield oed_source

    def save_config(self, filepath):
        """
        save data to directory, loadable later on
        Args:
            filepath (str): path to save the config file
        """
        Path(filepath).parents[0].mkdir(parents=True, exist_ok=True)
        with open(filepath, 'w') as info_file:
            json.dump(self.info, info_file, indent='  ', default=str)

    def save(self, path, version_name=None, compression=None, save_config=False):
        """
        helper function to save all OED data to a location with specific compression
        Args:
            version_name (str): use as a prefix for the file names, if None name will be taken from the cur_version file name
                          if it's a filepath
            path (str): output folder
            compression (str): type of compression to use
            save_config (bool): if true save the Exposure config as json
        """
        self.working_dir = path

        for oed_source in self.get_oed_sources():
            if version_name is None and oed_source.sources[oed_source.cur_version_name]['source_type'] == 'filepath':
                oed_name = Path(oed_source.sources[oed_source.cur_version_name]['filepath']).name
                if oed_source.cur_version_name.rsplit('_', 1)[-1] in PANDAS_COMPRESSION_MAP:
                    saved_version_name = oed_source.cur_version_name.rsplit('_', 1)[0]
                else:
                    saved_version_name = oed_source.cur_version_name
            elif version_name:
                oed_name = f'{version_name}_{OED_TYPE_TO_NAME[oed_source.oed_type]}'
                saved_version_name = version_name
            else:
                oed_name = f'{OED_TYPE_TO_NAME[oed_source.oed_type]}'
                saved_version_name = ''

            if save_config:  # we store filepath relative to config file
                filepath = Path(oed_name)
            else:
                filepath = Path(path, oed_name)

            if compression is None:
                if oed_source.sources[oed_source.cur_version_name]['source_type'] == 'filepath':
                    compression = oed_source.sources[oed_source.cur_version_name].get('extension')
                if compression is None:
                    compression = 'csv'

            filepath = filepath.with_suffix(PANDAS_COMPRESSION_MAP[compression])

            oed_source.save(saved_version_name + '_' + f'{compression}',
                            {'source_type': 'filepath', 'filepath': filepath, 'extension': compression})
        if save_config:
            self.save_config(Path(path, self.DEFAULT_EXPOSURE_CONFIG_NAME))

    def check(self, validation_config=None):
        """
        check that all OED files respect rules related to an OedSchema

        Args:
            validation_config (list): list of validation to perform, if None the default validation list will be used

        Returns:
            list of invalid data where validation action was 'return'

        Raises:
            OdsException if some invalid data is found

        """
        if validation_config is None:
            validation_config = self.validation_config
        validator = Validator(self)
        return validator(validation_config)



"""
This package is in charge of converting relevant ods values such as TIV or policy term into other currencies
"""
import datetime
import json
import os
import pandas as pd
from pathlib import Path

from .common import OdsException

try:
    from forex_python.converter import CurrencyRates as BaseFxCurrencyRates
    from forex_python.converter import RatesNotAvailableError

    class FxCurrencyRates(BaseFxCurrencyRates):
        """
        subclass of forex_python.converter.CurrencyRates that allow to set date to retrieve the rate from in the constructor
        """

        def __init__(self, *args, date_obj=None, **kwargs):
            super().__init__(*args, **kwargs)
            self.date_obj = date_obj

        def get_rate(self, base_cur, dest_cur, date_obj=None) -> float:
            """
            get rate from forex for curency pair (base_cur dest_cur)
            Args:
                base_cur (str): currency from
                dest_cur (str): currency to
                date_obj (datetime): if not None get the rate at this datetime instead of the latest forex date

            Returns:
                rate of exchange (float)

            Raises:
                OdsException if rate not found
            """
            if date_obj is None:
                date_obj = self.date_obj
            try:
                return super().get_rate(base_cur, dest_cur, date_obj)
            except RatesNotAvailableError as e:
                raise OdsException("Issue retrieving rate, most probably the forex api is down") from e

except ImportError:
    FxCurrencyRates = None

import logging

logger = logging.getLogger(__name__)

CURRENCY_COLUMN = {'Loc': 'LocCurrency',
                   'Acc': 'AccCurrency',
                   'ReinsInfo': 'ReinsCurrency'}


class DictBasedCurrencyRates:
    """
    Currency converter based on a dictionary of currency pair
    """

    def __init__(self, roe_dict):
        """
        Args:
            roe_dict: dict of currency pair tuple to rate of exchange
        """
        self.roe_dict = roe_dict

    @classmethod
    def from_dataframe(cls, df):
        """
        create DictBasedCurrencyRates from a DataFrame
        Args:
            df (pd.DataFrame): Dataframe with columns 'cur_from', 'cur_to', 'roe'

        Returns:
            DictBasedCurrencyRates
        """
        roe_dict = {}
        for row in df.to_dict(orient='records'):
            roe_dict[(row['cur_from'], row['cur_to'])] = row['roe']
        return cls(roe_dict)

    @classmethod
    def from_list(cls, currency_rates):
        """
        create DictBasedCurrencyRates from a list of (cur_from, cur_to, roe) triplet
        Args:
            currency_rates (list): list of (cur_from, cur_to, roe) triplet

        Returns:
            DictBasedCurrencyRates
        """
        return cls({(cur_from, cur_to): roe for cur_from, cur_to, roe in currency_rates})

    @classmethod
    def from_csv(cls, filepath_or_buffer, df_engine=pd, **kwargs):
        """
        create DictBasedCurrencyRates from csv file
        Args:
            filepath_or_buffer:  the csv file
            df_engine: engine that will read the csv file into a DataFrame
            **kwargs: extra named arg to passes to the read_csv function

        Returns:
            DictBasedCurrencyRates
        """
        if df_engine is None:
            raise Exception("df_engine parameter not specified, you must install pandas"
                            " or pass your DataFrame engine (modin, dask,...)")

        return cls.from_dataframe(df_engine.read_csv(filepath_or_buffer, **kwargs))

    @classmethod
    def from_parquet(cls, filepath_or_buffer, df_engine=pd, **kwargs):
        """
        create DictBasedCurrencyRates from parquet file
        Args:
            filepath_or_buffer: filepath or buffer of the csv file
            df_engine: engine that will read the parquet file into a DataFrame
            **kwargs: extra named arg to passes to the read_parquet function

        Returns:
            DictBasedCurrencyRates
        """
        if df_engine is None:
            raise Exception("df_engine parameter not specified, you must install pandas"
                            " or pass your DataFrame engine (modin, dask,...)")

        return cls.from_dataframe(df_engine.read_parquet(filepath_or_buffer, **kwargs))

    def get_rate(self, cur_from, cur_to):
        """
        get a rate of exchange to convert from cur_from to cur_to
        when currency pair (cur1, cur2) is requested
        it looks first for the pair
        if not found, look for (cur2, cur1) and return the inverse

        Args:
            cur_from (str): currency from
            cur_to (str): currency to

        Returns:
            rate of exchange (float)

        Raises:
            OdsException if currency pair not found
        """
        if (cur_from, cur_to) in self.roe_dict:
            return self.roe_dict[(cur_from, cur_to)]
        elif (cur_to, cur_from) in self.roe_dict:
            return 1. / self.roe_dict[(cur_to, cur_from)]
        else:
            raise OdsException(f"currency pair {(cur_from, cur_to)} is missing")


def create_currency_rates(currency_conversion):
    """
    Create a currency_rates object from currency_conversion dictionary or json file
    support:
        DictBasedCurrencyRates
        CurrencyRates from forex_python
    Args:
        currency_conversion (path_to_json or dict or None) : information to create the currency conversion object

    Returns:
        object implementing get_rate(self, cur_from, cur_to) method
    """

    def get_path(name):
        if Path(currency_conversion[name]).is_absolute():
            return currency_conversion[name]
        else:
            return os.path.join(str(currency_conversion['root_dir']), currency_conversion[name])

    if currency_conversion:
        if isinstance(currency_conversion, str):  # it is a json file
            root_dir = Path(currency_conversion).parents[0]
            with open(currency_conversion, 'r', encoding='utf-8') as f:
                currency_conversion = json.load(f)
            currency_conversion['root_dir'] = root_dir
        else:
            currency_conversion.setdefault('root_dir', Path.cwd())
    else:
        return None

    if currency_conversion.get('currency_conversion_type') == 'DictBasedCurrencyRates':
        if currency_conversion.get("source_type") == 'csv':
            return DictBasedCurrencyRates.from_csv(get_path('file_path'),
                                                   **currency_conversion.get("read_parameters", {}))
        elif currency_conversion.get("source_type") == 'parquet':
            return DictBasedCurrencyRates.from_parquet(get_path('file_path'),
                                                       **currency_conversion.get("read_parameters", {}))
        elif currency_conversion.get("source_type", '').lower() == 'dict':
            return DictBasedCurrencyRates(currency_conversion['currency_rates'])
        else:
            raise OdsException(
                f"Unsupported currency_conversion source type : {currency_conversion.get('source_type')}")
    elif currency_conversion.get('currency_conversion_type') == 'FxCurrencyRates':
        if FxCurrencyRates is None:
            raise OdsException(
                "You must install package forex-python to use builtin_currency_conversion_type FxCurrencyRates")

        _datetime = currency_conversion.get('datetime')
        if _datetime is not None:
            _datetime = datetime.datetime.fromisoformat(_datetime)
        return FxCurrencyRates(date_obj=_datetime, **currency_conversion.get("fx_currency_rates_parameters", {}))

    else:
        raise OdsException(
            f"unsupported currency_conversion_type {currency_conversion.get('currency_conversion_type')}")


def convert_currency(oed_df, oed_type, reporting_currency, currency_rate, oed_schema):
    """
    Convert inplace the columns in currency unit (BuildingTIV, LocNetPremium, LocDed1Building (depending on type),
    LocMinDed1Building, ...) to the reporting_currency
    store the original currency  in OriginCurency, and the rate of exchange in RateOfExchange

    Args:
        oed_df (pd.DataFrame): OED Dataframe (Loc, Acc, ...)
        oed_type (str): type of file (Loc, Acc, ...)
        reporting_currency (str): currency to convert to
        currency_rate: object to get the rate of exchange by calling get_rate(cur_from, cur_to)
        oed_schema (dict): ods_fields dict
    """
    if oed_type not in CURRENCY_COLUMN:
        return

    ods_fields = oed_schema.schema['input_fields'][oed_type]
    column_to_field = oed_schema.column_to_field(oed_df.columns, ods_fields)
    field_to_column = {field_info['Input Field Name']: column for column, field_info in column_to_field.items()}
    currency_col = field_to_column[CURRENCY_COLUMN[oed_type]]

    if 'OriginalCurrency' not in field_to_column:
        oed_df['OriginalCurrency'] = oed_df[currency_col]
        oed_df['RateOfExchange'] = 1.

    transaction_currencies = oed_df[currency_col].unique()

    if set(transaction_currencies) - {reporting_currency}:
        if currency_rate is None:
            raise OdsException(
                f'Currency Convertion needs to be specified in order to convert term to reporting currency {reporting_currency}')

    for orig_cur in transaction_currencies:
        if orig_cur == reporting_currency:
            continue
        rate = currency_rate.get_rate(orig_cur, reporting_currency)

        orig_cur_rows = (oed_df[currency_col] == orig_cur)
        oed_df.loc[orig_cur_rows, 'RateOfExchange'] *= rate
        for field, column in field_to_column.items():
            field_type = ods_fields[field.lower()].get('Back End DB Field Name', '').lower()
            if (field_type in ['tax', 'grosspremium', 'netpremium', 'brokerage', 'extraexpenselimit', 'minded',
                               'maxded']
                    or field.endswith('tiv')):
                row_filter = orig_cur_rows
            elif field_type == 'ded':
                column_type_name = field.replace('ded', 'dedtype')
                row_filter = orig_cur_rows & (oed_df[field_to_column[column_type_name]] == 0)
            elif field_type == 'limit':
                column_type_name = field.replace('limit', 'limittype')
                row_filter = orig_cur_rows & (oed_df[field_to_column[column_type_name]] == 0)
            elif field_type in ['payoutstart', 'payoutend', 'payoutlimit']:
                column_type_name = 'payouttype'
                row_filter = orig_cur_rows & (oed_df[field_to_column[column_type_name]] == 0)
            elif field_type in ['triggerstart', 'triggerend']:
                column_type_name = 'triggertype'
                row_filter = orig_cur_rows & (oed_df[field_to_column[column_type_name]] == 0)
            else:  # not a currency unit column we go to the next one
                continue
            oed_df.loc[row_filter, column] *= rate

    oed_df[currency_col] = reporting_currency
    oed_df[currency_col] = oed_df[currency_col].astype('category')



"""
common static variable and ods_tools exceptions
"""
from urllib.parse import urlparse
from pathlib import Path


class OdsException(Exception):
    pass


def is_relative(filepath):
    """
    return True is path is relative meaning it is neither internet RFC nor os absolute path
    Args:
        filepath (str: path est

    Returns:
        boolean
    """
    url_parsed = urlparse(str(filepath))
    return not (all([url_parsed.scheme, url_parsed.netloc]) or Path(filepath).is_absolute())


# PANDAS_COMPRESSION_MAP is also used to order the preferred input format in ExposureData.from_dir
PANDAS_COMPRESSION_MAP = {
    'parquet': '.parquet',
    'csv': '.csv',
    'zip': '.zip',
    'gzip': '.gz',
    'bz2': '.bz2',
    'zstd': '.zst',
}

PANDAS_DEFAULT_NULL_VALUES = {
    '-1.#IND',
    '1.#QNAN',
    '1.#IND',
    '-1.#QNAN',
    '#N/A N/A',
    '#N/A',
    'N/A',
    'n/a',
    'NA',
    '#NA',
    'NULL',
    'null',
    'NaN',
    '-NaN',
    '-NaN',
    'nan',
    '-nan',
    '',
}

USUAL_FILE_NAME = {
    'location': ['location'],
    'account': ['account'],
    'ri_info': ['ri_info', 'reinsinfo'],
    'ri_scope': ['ri_scope', 'reinsscope'],
}

OED_TYPE_TO_NAME = {
    'Loc': 'location',
    'Acc': 'account',
    'ReinsInfo': 'ri_info',
    'ReinsScope': 'ri_scope'
}

OED_NAME_TO_TYPE = {value: key for key, value in OED_TYPE_TO_NAME.items()}

OED_IDENTIFIER_FIELDS = {
    'Loc': ['PortNumber', 'AccNumber', 'LocNumber'],
    'Acc': ['PortNumber', 'AccNumber', 'PolNumber', 'LayerNumber'],
    'ReinsInfo': ['ReinsNumber', 'ReinsLayerNumber', 'ReinsName', 'ReinsPeril'],
    'ReinsScope': ['ReinsNumber', 'PortNumber', 'AccNumber', 'LocNumber']
}

VALIDATOR_ON_ERROR_ACTION = {'raise', 'log', 'ignore', 'return'}
DEFAULT_VALIDATION_CONFIG = [
    {'name': 'source_coherence', 'on_error': 'raise'},
    {'name': 'required_fields', 'on_error': 'raise'},
    {'name': 'unknown_column', 'on_error': 'log'},
    {'name': 'valid_values', 'on_error': 'raise'},
    {'name': 'perils', 'on_error': 'raise'},
    {'name': 'occupancy_code', 'on_error': 'raise'},
    {'name': 'construction_code', 'on_error': 'raise'},
    {'name': 'country_and_area_code', 'on_error': 'raise'},
]

OED_PERIL_COLUMNS = ['AccPeril', 'PolPerilsCovered', 'PolPeril', 'CondPeril', 'LocPerilsCovered', 'LocPeril',
                     'ReinsPeril']



import functools
import json
import logging

import numpy as np
import pandas as pd
from pathlib import Path
from collections.abc import Iterable

from .common import (OdsException, OED_PERIL_COLUMNS, OED_IDENTIFIER_FIELDS, DEFAULT_VALIDATION_CONFIG,
                     VALIDATOR_ON_ERROR_ACTION)
from .schema import OedSchema

logger = logging.getLogger(__name__)


class Validator:
    def __init__(self, exposure):
        """
        create a Validator object for exposure data
        Args:
            exposure: OedExposure object
        """
        self.exposure = exposure

        self.column_to_field_maps = {}
        self.identifier_field_maps = {}
        self.field_to_column_maps = {}
        for oed_source in exposure.get_oed_sources():
            self.column_to_field_maps[oed_source] = oed_source.get_column_to_field()
            self.identifier_field_maps[oed_source] = [column for column, field_info in self.column_to_field_maps[oed_source].items()
                                                      if field_info['Input Field Name'] in OED_IDENTIFIER_FIELDS[oed_source.oed_type]]
            field_to_column = {}
            self.field_to_column_maps[oed_source] = field_to_column
            for column, field_info in self.column_to_field_maps[oed_source].items():
                if field_info['Input Field Name'] in field_to_column:
                    if field_info['Input Field Name'].endswith('XX') or field_info['Input Field Name'].endswith('ZZZ'):
                        if not isinstance(field_to_column[field_info['Input Field Name']], list):
                            field_to_column[field_info['Input Field Name']] = [field_to_column[field_info['Input Field Name']]]
                        field_to_column[field_info['Input Field Name']].append(column)

                    else:
                        raise OdsException(f"Oed file {oed_source.oed_name}, {oed_source.current_source}"
                                           f" contain multiple instances of unique field {field_info['Input Field Name']}"
                                           f" {field_to_column[field_info['Input Field Name']]} and {column}")
                else:
                    field_to_column[field_info['Input Field Name']] = column

    def __call__(self, validation_config):
        """
        run all check from validation_config
        Args:
            validation_config = list of checks to perform with their action
                - ex [{'name': 'required_fields', 'on_error': 'raise'}, ...]

        Returns:
            list of errors from check with on_error "return"
        """
        if validation_config is None:
            validation = DEFAULT_VALIDATION_CONFIG
        elif isinstance(validation_config, Iterable):
            validation = validation_config
        elif isinstance(validation_config, [str, Path]):
            validation = json.load(validation_config)
        else:
            raise OdsException("Unsupported validation type")

        invalid_data_group = {}

        for check in validation:
            check_fct = getattr(self, 'check_' + str(check['name']), None)
            if check.get('on_error') not in VALIDATOR_ON_ERROR_ACTION:
                raise OdsException('Unknown check on_error action' + str(check.get('on_error')))
            if check['on_error'] == 'ignore':
                continue
            if hasattr(check_fct, '__call__'):
                invalid_data_group.setdefault(check['on_error'], []).extend(check_fct())
            else:
                raise OdsException('Unknown check name ' + str(check['name']))

        raise_msg = invalid_data_group.get('raise', [])
        log_msg = invalid_data_group.get('log', [])
        return_msg = invalid_data_group.get('return', [])

        def invalid_data_to_str(_data):
            return f"in {_data['name']} {_data['source']}\n {_data['msg']}"

        for invalid_data in log_msg:
            logger.info(invalid_data_to_str(invalid_data))
        if raise_msg:
            raise OdsException('\n'.join(invalid_data_to_str(invalid_data) for invalid_data in raise_msg))
        return return_msg

    def check_source_coherence(self):
        """"""
        invalid_data = []
        if not self.exposure.location:
            invalid_data.append({'name': 'location', 'source': None,
                                 'msg': f"Exposure needs a Location file, location={self.exposure.location}"})

        if self.exposure.ri_info or self.exposure.ri_scope:
            if not self.exposure.account:
                invalid_data.append({'name': 'account', 'source': None,
                                     'msg': f"Exposure needs account if reinsurance is provided account={self.exposure.account}"})

            if not self.exposure.ri_info and self.exposure.ri_scope:
                invalid_data.append({'name': 'reinsurance', 'source': None,
                                     'msg': f"Exposure needs both ri_scope and ri_scope for reinsurance"
                                            f"ri_info={self.exposure.ri_info} ri_scope={self.exposure.ri_scope}"})
        return invalid_data

    def check_required_fields(self):
        """
        using Oed input_field definition, check all require field
        error is raised if
         - required column is missing
         - value is missing and "Allow blanks?" == 'NO'
        Returns:
            list of invalid_data
        """

        invalid_data = []
        for oed_source in self.exposure.get_oed_sources():
            input_fields = oed_source.get_input_fields()
            identifier_field = self.identifier_field_maps[oed_source]
            field_to_columns = self.field_to_column_maps[oed_source]

            for field_info in input_fields.values():
                if field_info['Input Field Name'] not in field_to_columns:
                    if field_info.get('Required Field') == 'R':
                        invalid_data.append({'name': oed_source.oed_name, 'source': oed_source.current_source,
                                             'msg': f"missing required column {field_info['Input Field Name']}"})
                    continue
                columns = field_to_columns[field_info['Input Field Name']]
                if isinstance(columns, str):
                    columns = [columns]
                for column in columns:
                    if (field_info.get("Allow blanks?").upper() == 'NO'
                            and (oed_source.dataframe[column].isna().any()
                                 or (field_info['pd_dtype'] == 'str' and oed_source.dataframe[column].isnull().any()))):
                        invalid_data.append({'name': oed_source.oed_name, 'source': oed_source.current_source,
                                             'msg': f"column '{column}' has missing values in \n"
                                                    f"{oed_source.dataframe[identifier_field + [column]]}"})
        return invalid_data

    def check_unknown_column(self):
        """
        using Oed input_field definition, check that all column are OED column
        Returns:
            list of invalid_data
        """

        invalid_data = []
        for oed_source in self.exposure.get_oed_sources():
            column_to_field = self.column_to_field_maps[oed_source]
            for column in oed_source.dataframe.columns:
                if column not in column_to_field:
                    invalid_data.append({'name': oed_source.oed_name, 'source': oed_source.current_source,
                                         'msg': f"column '{column}' is not a valid oed field"})
        return invalid_data

    def check_valid_values(self):
        """
        using Oed input_field definition, check that values are valid for field that define a 'Valid value range'
        Returns:
            list of invalid_data
        """
        invalid_data = []
        for oed_source in self.exposure.get_oed_sources():
            column_to_field = self.column_to_field_maps[oed_source]
            identifier_field = self.identifier_field_maps[oed_source]
            for column, field_info in column_to_field.items():
                valid_ranges = field_info['Valid value range']
                if valid_ranges != 'n/a':
                    is_valid_value = functools.partial(OedSchema.is_valid_value,
                                                       valid_ranges=valid_ranges,
                                                       allow_blanks=field_info['Allow blanks?'].lower() == 'yes')
                    invalid_range_data = oed_source.dataframe[~oed_source.dataframe[column].apply(is_valid_value)]
                    if not invalid_range_data.empty:
                        invalid_data.append({'name': oed_source.oed_name, 'source': oed_source.current_source,
                                             'msg': f"column '{column}' has values outside range.\n"
                                                    f"{invalid_range_data[identifier_field + [column]]}"})
        return invalid_data

    def check_perils(self):
        """
        using Oed perils list specification, check that Peril column have valid peril values
        Returns:
            list of invalid_data
        """
        invalid_data = []
        for oed_source in self.exposure.get_oed_sources():
            identifier_field = self.identifier_field_maps[oed_source]
            for column in oed_source.dataframe.columns.intersection(set(OED_PERIL_COLUMNS)):
                peril_values = oed_source.dataframe[column].str.split(';').apply(pd.Series, 1).stack()
                invalid_perils = oed_source.dataframe.iloc[
                    peril_values[~peril_values.isin(self.exposure.oed_schema.schema['perils']['info'])].index.droplevel(-1)]
                if not invalid_perils.empty:
                    invalid_data.append({'name': oed_source.oed_name, 'source': oed_source.current_source,
                                         'msg': f"{column} has invalid perils.\n"
                                                f"{invalid_perils[identifier_field + [column]]}"})
        return invalid_data

    def check_occupancy_code(self):
        """
        using Oed occupancy_code list specification, check that occupancy_code column have valid occupancy_code values
        Returns:
            list of invalid_data
        """
        invalid_data = []
        for oed_source in self.exposure.get_oed_sources():
            occupancy_code_column = self.field_to_column_maps[oed_source].get('occupancycode')
            if occupancy_code_column is None:
                continue
            identifier_field = self.identifier_field_maps[oed_source]
            invalid_occupancy_code = oed_source.dataframe[
                ~oed_source.dataframe[occupancy_code_column].astype(str).isin(self.exposure.oed_schema.schema['occupancy'])]
            if not invalid_occupancy_code.empty:
                invalid_data.append({'name': oed_source.oed_name, 'source': oed_source.current_source,
                                     'msg': f"invalid OccupancyCode.\n"
                                            f"{invalid_occupancy_code[identifier_field + [occupancy_code_column]]}"})
        return invalid_data

    def check_construction_code(self):
        """
        using Oed occupancy_code list specification, check that occupancy_code column have valid occupancy_code values
        Returns:
            list of invalid_data
        """
        invalid_data = []
        for oed_source in self.exposure.get_oed_sources():
            construction_code_column = self.field_to_column_maps[oed_source].get('constructioncode')
            if construction_code_column is None:
                continue
            identifier_field = self.identifier_field_maps[oed_source]
            invalid_construction_code = oed_source.dataframe[
                ~oed_source.dataframe[construction_code_column].astype(str).isin(self.exposure.oed_schema.schema['construction'])]
            if not invalid_construction_code.empty:
                invalid_data.append({'name': oed_source.oed_name, 'source': oed_source.current_source,
                                     'msg': f"invalid ConstructionCode.\n"
                                            f"{invalid_construction_code[identifier_field + [construction_code_column]]}"})
        return invalid_data

    def check_country_and_area_code(self):
        """
        using Oed country_and_area_code list specification,
        check that country and area_code column have valid country or (country and area_code) pair values
        Returns:
            list of invalid_data
        """
        invalid_data = []
        for oed_source in self.exposure.get_oed_sources():
            country_code_column = self.field_to_column_maps[oed_source].get('countrycode')
            if country_code_column is None:
                continue
            identifier_field = self.identifier_field_maps[oed_source]
            area_code_column = self.field_to_column_maps[oed_source].get('areacode')
            if area_code_column is not None:
                country_only_df = oed_source.dataframe[oed_source.dataframe[area_code_column].isin([np.nan, ''])]
                country_area_df = oed_source.dataframe[~oed_source.dataframe[area_code_column].isin([np.nan, ''])]
                invalid_country_area = (country_area_df[
                    ~(country_area_df[[country_code_column, area_code_column]]
                      .apply(tuple, axis=1)
                      .isin(self.exposure.oed_schema.schema['country_area'])
                      )]
                )
                if not invalid_country_area.empty:
                    invalid_data.append({'name': oed_source.oed_name, 'source': oed_source.current_source,
                                         'msg': f"invalid countrycode areacode pair.\n"
                                                f"{invalid_country_area[identifier_field + [country_code_column, area_code_column]]}"})
            else:
                country_only_df = oed_source.dataframe
            invalid_country = (country_only_df[~country_only_df[country_code_column]
                                               .isin(set(self.exposure.oed_schema.schema['country']) | {np.nan, ''})])
            if not invalid_country.empty:
                invalid_data.append({'name': oed_source.oed_name, 'source': oed_source.current_source,
                                     'msg': f"invalid countrycode.\n"
                                            f"{invalid_country[identifier_field + [country_code_column]]}"})
        return invalid_data



import io
import os
import re
import setuptools


SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))


def get_readme():
    try:
        with io.open(os.path.join(SCRIPT_DIR, 'ods_tools', 'README.md'), encoding='utf-8') as readme:
            return readme.read()
    except FileNotFoundError:
        with io.open(os.path.join(SCRIPT_DIR, 'README.md'), encoding='utf-8') as readme:
            return readme.read()


def get_version():
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    with io.open(os.path.join(SCRIPT_DIR, 'ods_tools', '__init__.py'), encoding='utf-8') as init_py:
        return re.search('__version__ = [\'"]([^\'"]+)[\'"]', init_py.read()).group(1)


def get_install_requirements():
    with io.open(os.path.join(SCRIPT_DIR, 'requirements.in'), encoding='utf-8') as reqs:
        return reqs.readlines()


version = get_version()
readme = get_readme()
reqs = get_install_requirements()

setuptools.setup(
    name="ods_tools",
    version=version,
    include_package_data=True,
    package_data={
        "": ["*.md"],                # Copy in readme
        "ods_tools": ["data/*"]      # Copy spec JSON/CSV
    },
    entry_points={
        'console_scripts': [
            'ods_tools=ods_tools.main:main',
        ]
    },
    author='Oasis LMF',
    author_email="support@oasislmf.org",
    packages=['ods_tools', 'ods_tools.oed', 'ods_tools.data'],
    package_dir={'ods_tools': 'ods_tools'},
    python_requires='>=3.7',
    install_requires=reqs,
    description='Tools to manage ODS files',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/OasisLMF/OpenDataStandards',
)



"""
suite of tools to read and write ODS file in csv of parquet
"""
__all__ = [
    'main',
    'convert',
    'check'
]

import argparse
import logging

from .oed import OedExposure, OdsException

logger = logging.getLogger(__name__)


def get_oed_exposure(config_json=None, oed_dir=None, **kwargs):
    if config_json:
        return OedExposure.from_config(config_json, **kwargs)
    elif oed_dir:
        return OedExposure.from_dir(oed_dir, **kwargs)
    else:
        return OedExposure(**kwargs)


def extract_exposure_args(kwargs):
    exposure_kwargs = {}
    for param in ['location', 'account', 'ri_info', 'ri_scope', 'oed_dir', 'check_oed', 'config_json', 'validation_config']:
        value = kwargs.pop(param, None)
        if value is not None:
            exposure_kwargs[param] = value
    return exposure_kwargs


def check(**kwargs):
    """run the check command on Exposure"""
    oed_exposure = get_oed_exposure(**extract_exposure_args(kwargs))
    oed_exposure.check(**kwargs)


def convert(**kwargs):
    """Convert exposure data to an other format (ex: csv to parquet)"""
    path = kwargs.pop('output_dir', None) or kwargs.get('oed_dir', None)
    if not path:
        raise OdsException('--output-dir or --oed-dir need to be provided to perform convert')
    oed_exposure = get_oed_exposure(**extract_exposure_args(kwargs))
    oed_exposure.save(path=path, **kwargs)


command_action = {
    'check': check,
    'convert': convert,
}


def add_exposure_data_args(command):
    command.add_argument('--location', help='Path to the location file', default=None)
    command.add_argument('--account', help='Path to the account file', default=None)
    command.add_argument('--ri-info', help='Path to the ri_info file', default=None)
    command.add_argument('--ri-scope', help='Path to the ri_scope file', default=None)
    command.add_argument('--oed-dir', help='Path to the OED directory containing stardard named OED files', default=None)
    command.add_argument('--config-json', help='Path to the config_json file', default=None)
    command.add_argument('--validation-config', help='Path to the validation_config file', default=None)


main_parser = argparse.ArgumentParser()

oed_exposure_creation = """
there are several options to specify the exposure data,
 - by providing  the path to each OED source using `--location`, `--account`, `--ri-info`, `--ri-scope`.
 - by providing  an OED config json file using `--config-json`.
 - by providing  the path to the directory where the exposure is stored using `--oed-dir`.
 
if multiple options are use at the same time, --config-json will have the priority over --oed-dir
specific paths (--location, --account, --ri-info, --ri-scope) will overwrite the path found in (--config-json or --oed-dir)
 """


convert_description = """
convert OED files to an other format
"""

command_parser = main_parser.add_subparsers(help='command [convert]', dest='command', required=True)
convert_command = command_parser.add_parser('convert', description=convert_description +
                                            oed_exposure_creation, formatter_class=argparse.RawTextHelpFormatter)
add_exposure_data_args(convert_command)
convert_command.add_argument('--check-oed', help='if True, OED file will be checked before convertion', default=False)
convert_command.add_argument('--output-dir', help='path of the output directory', required=False)
convert_command.add_argument('-c', '--compression', help='compression to use (ex: parquet, zip, gzip, csv,...)', required=True)
convert_command.add_argument('--save-config', help='if True, OED config file will be save in the --path directory', default=False)
convert_command.add_argument('-v', '--logging-level', help='logging level (debug:10, info:20, warning:30, error:40, critical:50)',
                             default=30, type=int)

check_description = """
check exposure data.
"""

check_command = command_parser.add_parser('check', description=check_description + oed_exposure_creation,
                                          formatter_class=argparse.RawTextHelpFormatter)
add_exposure_data_args(check_command)
check_command.add_argument('-v', '--logging-level', help='logging level (debug:10, info:20, warning:30, error:40, critical:50)',
                           default=30, type=int)


def main():
    """command line interface for ODS conversion between csv and parquet"""
    kwargs = vars(main_parser.parse_args())
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    command_action[kwargs.pop('command')](** kwargs)


if __name__ == "__main__":
    main()
